from . import logs

import json
import tempfile
import subprocess
import shlex
import pathlib

import pdal


def run(command):
    args = shlex.split(command)
    p = subprocess.Popen(args,
                            stdin = subprocess.PIPE,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE,)
    ret = p.communicate()

    if p.returncode != 0:
        error = ret[1].decode('utf-8','replace')
        raise RuntimeError(error)

    response = ret[0].decode('utf-8','replace')
    return response

class Product(object):
    def __init__(self, reader):
        self.reader = reader
        self.path = pathlib.Path(tempfile.NamedTemporaryFile(suffix='.tif', delete=False).name)
        self.validate()

    def __del__(self):
        self.path.unlink()

    def getStage(self):
        pass

    def validate(self):
        pass

    def process(self):
        stage = self.getStage()

        pipeline = self.reader.get() | stage

        count = 0
        if pipeline.streamable:
            count = pipeline.execute_streaming(chunk_size=self.reader.args.chunk_size)
        else:
            count = pipeline.execute()
        logs.logger.info(f'Wrote {self.name} product for {count} points')

class Intensity(Product):
    def __init__(self, reader):
        self.name = 'Intensity'
        super().__init__(reader)


    def getStage(self):
        stage = pdal.Writer.gdal(
            filename=str(self.path),
            data_type='uint16_t',
            dimension='Intensity',
            output_type = 'idw',
            resolution=self.reader.args.resolution,
        )

        return stage

    def validate(self):
        qi = self.reader.get().quickinfo
        for key in qi:
            dimensions = [i.strip() for i in qi[key]['dimensions'].split(',')]
            if self.name not in dimensions:
                raise RuntimeError(f"{self.name} information not available, this tool cannot run")


class DSM(Product):
    def __init__(self, reader):
        self.name = 'DSM'
        super().__init__(reader)


    def getStage(self):
        stage = pdal.Writer.gdal(filename=str(self.path),
                                 data_type='float',
                                 dimension='Z',
                                 output_type = 'idw',
                                 resolution=self.reader.args.resolution)
        return stage

class Daylight(object):
    def __init__(self, dsm: DSM):
        self.name = 'Daylight'        
        self.path = pathlib.Path(tempfile.NamedTemporaryFile(suffix='.tif', delete=False).name)
        self.dsm = dsm

    def getImageCenter(self):
        # Run our pipeline

        command = f'gdalinfo -json {self.dsm.path}'
        response = run(command)
        j = json.loads(response)
        corner = j['wgs84Extent']['coordinates'][0][0]
        logs.logger.info(f'Fetched image center {corner}')
        return corner


    def process(self):

        lng, lat = self.getImageCenter()

        command = f"""whitebox_tools -r=TimeInDaylight  \
        -i {self.dsm.path} -o {self.path}  --az_fraction=15.0 \
        --max_dist=100.0 --lat={lat:.5f} --long={lng:.5f} """
        logs.logger.info(f"Processing ambient occlusion '{command}'")
        response = run(command)
        logs.logger.info(f"Processed ambient occlusion ")


class Reader(object):
    def __init__(self, args ):
        self.args = args
        self.name = 'Reader'

        self.reader_args = None
        if 'reader_args' in self.args:
            with open(self.args.reader_args,'r') as f:
                self.reader_args = json.loads(f.read())

        if '.json' in self.args.input.suffixes:
            self.inputType = 'pipeline'
        else:
            self.inputType = 'readable'

    def get(self):
        if self.inputType == 'pipeline':
            reader = self.readPipeline()
        else:
            reader = readFile()

        return reader
    def readFile(self):
        reader = pdal.Reader(str(self.args.input), *self.reader_args)
        pipeline = reader.pipeline()
        breakpoint()
        return pipeline

    def readPipeline(self):
        if self.inputType != 'pipeline':
            raise RuntimeError("Data type is not pipeline!")
        j = self.args.input.read_bytes().decode('utf-8')
        stages = pdal.pipeline._parse_stages(j)
        p = pdal.Pipeline(stages)

        # strip off any writers we're making our own
        stages = []
        for stage in p.stages:
            if stage.type.split('.')[0] != 'writers':
                stages.append(stage)

        p = pdal.Pipeline(stages)
        return p







from . import data
from . import blend
from .logs import logger 


def doIt(args):
    logger.info(f"opening {args.input} for processing")
    reader = data.Reader(args)

    intensity = data.Intensity(reader)
    dsm = data.DSM(reader)

    dsm.process()
    intensity.process()

    daylight = data.Daylight(dsm)
    daylight.process()

    b = blend.Blend(intensity, dsm, daylight)
    b.write(args)


from . import data
from . import blend
from .logs import logger
import uuid

import numpy as np
import pdal
import requests
import pathlib
import geocoder
import pyproj
from osgeo import gdal

from shapely.geometry import shape, Polygon, LineString
from shapely.ops import transform
from shapely import wkt



def get_polyline(args):

    line_dd = wkt.loads(args.line)
    line_srs = pyproj.CRS("EPSG:4326")
    if args.line_srs:
        line_srs = pyproj.CRS(args.line_srs)

    dsm_srs_wkt = gdal.Open(f'{args.input}').GetSpatialRef().ExportToWkt()
    if not dsm_srs_wkt:
        raise Exception("Input does not have a SRS, we have to stop. Use a VRT to define it or something")
    dsm_srs = pyproj.CRS(dsm_srs_wkt)

    forward = pyproj.Transformer.from_crs(line_srs, dsm_srs, always_xy=True).transform
    reverse = pyproj.Transformer.from_crs(dsm_srs, line_srs, always_xy=True).transform

    line_wm = transform(forward, line_dd)
    poly_wm = line_wm.buffer(2*args.range, cap_style='square')


    poly_dd = transform(reverse, poly_wm)
    return line_dd, line_wm, poly_dd, poly_wm




def do(args):
    logger.info(f"opening {args.input} for processing")


    # python do.py "MULTILINESTRING ((-91.563762 41.649644,-91.5638203 41.6496412,-91.563971 41.649634,-91.564446 41.649632,-91.5648984 41.64963,-91.565474 41.649639,-91.566724 41.649644))"
    line_dd, line_wm, poly_dd, poly_wm = get_polyline(args)

    raster = gdal.Open(f'{args.input}')

    dense = densify(line_wm, 10)
    points = dense.xy

    sheds = []
    for i in range(len(points[0])):

        x0 = points[0][i]
        y0 = points[1][i]

        sheds.append(viewshed(args,raster, x0, y0))


    options = gdal.BuildVRTOptions(resampleAlg='nearest', addAlpha=True,srcNodata=0)
    vrt = gdal.BuildVRT('/vsimem/my.vrt', [r for r in sheds], options=options)
    vrt = None

    driver = gdal.GetDriverByName('GTiff')
    l = gdal.ReadDir('/vsimem/')
    vs = gdal.Open('/vsimem/my.vrt')
    dst_ds = driver.CreateCopy(f'{args.output}', vs, strict=0)


def compute_utmzone(bounds):
    from pyproj import CRS
    from pyproj.aoi import AreaOfInterest
    from pyproj.database import query_utm_crs_info

    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=bounds[0],
            south_lat_degree=bounds[1],
            east_lon_degree=bounds[2],
            north_lat_degree=bounds[3],
        ),
    )
    utm_crs = CRS.from_epsg(utm_crs_list[0].code)
    return utm_crs


def densify(line_geometry, step):
    length_m = line_geometry.length  # get the length
    xy = []  # to store new tuples of coordinates

    for distance_along_old_line in np.arange(0, int(length_m), step):
        point = line_geometry.interpolate(
            distance_along_old_line
        )  # interpolate a point every step along the old line
        xp, yp = point.x, point.y  # extract the coordinates

        xy.append((xp, yp))  # and store them in xy list

    new_line = LineString(
        xy
    )  # Here, we finally create a new line with densified points.

    return new_line


def warp(raster_filename, outSrs, inSrs=None):
    pass


def viewshed(args, raster_ds, x, y, rng=500):

    band = raster_ds.GetRasterBand(1)
    filename = f'/vsimem/{uuid.uuid4()}.tif'
    ds = gdal.ViewshedGenerate(
        raster_ds.GetRasterBand(1),
        "GTiff",
        filename,
        ["INTERLEAVE=BAND"],
        x,
        y,
        1.5,
        args.height,  # targetHeight
        255,  # visibleVal
        0,  # invisibleVal
        0,  # outOfRangeVal
        -1.0,  # noDataVal,
        0.85714,  # dfCurvCoeff
        gdal.GVM_Edge,
        rng,  # maxDistance
        heightMode=gdal.GVOT_NORMAL,
        options=["UNUSED=YES"],
    )

    del ds

    return filename

